__all__ = ["patch", "__package__"]

from . import importer as _importer
from . import paths as _paths
from .patch import patch

_PACKAGE_ROOT_PATH = _paths.find_package_root()
# In order for relative imports to work, we need to know the __package__ name of the
# notebook. This is imported using %run in notebooks
__package__ = _importer.determine_package_name(_PACKAGE_ROOT_PATH)

# Add root package to sys path & install import hook
_importer.install_hook(_PACKAGE_ROOT_PATH)



from typing import Callable, Type, TypeVar

T = TypeVar("T")


def patch(cls: Type) -> Callable[[T], T]:
    """Decorator to monkey-patch additional methods to a class.

    At import-time, this will disappear and the source code itself will be transformed
    Inside notebooks, the implementation below will be used.

    :param cls:
    :return:
    """

    def _notebook_patch_impl(func):
        setattr(cls, func.__name__, func)
        return func

    return _notebook_patch_impl



import os
import pathlib
from typing import AnyStr, Union

ROOT_PACKAGE_FILES = ["pyproject.toml", "setup.py"]


def find_package_root(path: Union[AnyStr, os.PathLike] = None) -> pathlib.Path:
    """Find the root path of the repository.
    This assumes that all notebooks are imported
    relative to this directory.

    :param path: starting path to traverse from
    :return:
    """
    if path is None:
        path = pathlib.Path.cwd()
    else:
        path = pathlib.Path(path).absolute()

    while path != path.parent:
        for p in ROOT_PACKAGE_FILES:
            if (path / p).exists():
                return path
        path = path.parent
    raise ValueError("Could not find package root")



import argparse
import logging
import pathlib
from concurrent.futures import ProcessPoolExecutor

import nbclient
import nbformat

logger = logging.getLogger(__name__)


def run_notebook(path: pathlib.Path, cell_timeout: int = 120):
    logger.debug(f"Executing {path}")
    nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
    client = nbclient.NotebookClient(
        nb, timeout=cell_timeout, resources={"metadata": {"path": path.parent}}
    )
    client.execute()
    logger.debug(f"Finished executing {path}")


def configure(subparsers):
    parser = subparsers.add_parser(
        "test",
        description="Test literary notebooks in parallel",
    )
    parser.add_argument(
        "source", type=pathlib.Path, help="source directory for notebooks"
    )
    parser.add_argument(
        "-j", "--jobs", default=None, type=int, help="number of " "parallel jobs to run"
    )
    return parser


def run(args):
    executor = ProcessPoolExecutor(max_workers=args.jobs)
    executor.map(run_notebook, args.source.glob("*.ipynb"))
    executor.shutdown()



import pathlib

from .. import package


def configure(subparsers):
    parser = subparsers.add_parser(
        "build",
        description="Build a pure-Python package from a set of Jupyter notebooks",
    )
    parser.add_argument(
        "source", type=pathlib.Path, help="source directory for notebooks"
    )
    parser.add_argument(
        "dest_root", type=pathlib.Path, help="path to parent of generated package"
    )
    return parser


def run(args):
    package.build_package(args.source, args.dest_root)



import ast
import logging

import astunparse
import traitlets.config
from jinja2 import DictLoader
from nbconvert import exporters
from traitlets import default, import_item, List

from .preprocessor import LiteraryPythonPreprocessor
from .utils import escape_triple_quotes

logger = logging.getLogger(__name__)


class ASTUnparser(astunparse.Unparser):
    """AST unparser with additional preference for triple-quoted multi-line strings"""

    def _Constant(self, tree):
        if isinstance(tree.value, str) and "\n" in tree.value:
            self.write(f'"""{escape_triple_quotes(tree.value)}"""')
            return

        super()._Constant(tree)


# Monkey patch to ensure correctness
astunparse.Unparser = ASTUnparser


class LiteraryPythonExporter(exporters.PythonExporter):
    def __init__(self, *args, **kwargs):
        """
        Public constructor

        Parameters
        ----------
        config : ``traitlets.config.Config``
            User configuration instance.
        `**kw`
            Additional keyword arguments passed to parent __init__

        """

        super().__init__(*args, **kwargs)

        self._init_transformers()

    transformers = List(["literary.transformers.PatchTransformer"]).tag(config=True)

    def _init_transformers(self):
        self._transformers = []

        for value in self.transformers:
            if isinstance(value, str):
                value = import_item(value)
            self._transformers.append(value())

    @default("template_file")
    def _template_file_default(self):
        return "literary"

    @default("extra_loaders")
    def _extra_loaders_default(self):
        loader = DictLoader(
            {
                "literary": '''
{%- extends 'python/index.py.j2' -%}

{% block header %}
"""
{%- for cell in nb.cells -%}
{%- if "docstring" in cell.metadata.tags -%}
{{ cell.source | escape_triple_quotes }}
{% endif -%}
{%- endfor -%}
"""
{% endblock header %}

{%- block any_cell scoped -%}
{%- if "export" in cell.metadata.tags -%}
{{ super() }}
{%- endif -%}
{%- endblock any_cell -%}

                             '''
            }
        )
        return [loader]

    @default("default_preprocessors")
    def _default_preprocessors_default(self):
        return [LiteraryPythonPreprocessor]

    @default("exclude_input_prompt")
    def _exclude_input_prompt_default(self):
        return True

    @property
    def default_config(self):
        c = traitlets.config.Config(
            {
                "LiteraryPythonPreprocessor": {"enabled": True},
            }
        )
        c.merge(super().default_config)
        return c

    def default_filters(self):
        yield from super().default_filters()
        yield "escape_triple_quotes", escape_triple_quotes

    def from_notebook_node(self, nb, resources=None, **kwargs):
        body, resources = super().from_notebook_node(nb, resources, **kwargs)
        node = ast.parse(body)

        # Support ast.NodeTransformer and custom transformers using
        # `visit()` API
        for transformer in self._transformers:
            node = transformer.visit(node)

        return astunparse.unparse(node), resources
